@@ -67,103 +67,6 @@ command_table_create_parse_flags(grn_ctx *ctx,
   return flags;
 }
 
-static grn_bool
-grn_proc_table_set_token_filters_put(grn_ctx *ctx,
-                                     grn_obj *token_filters,
-                                     const char *token_filter_name,
-                                     int token_filter_name_length)
-{
-  grn_obj *token_filter;
-
-  token_filter = grn_ctx_get(ctx,
-                             token_filter_name,
-                             token_filter_name_length);
-  if (token_filter) {
-    GRN_PTR_PUT(ctx, token_filters, token_filter);
-    return GRN_TRUE;
-  } else {
-    GRN_PLUGIN_ERROR(ctx,
-                     GRN_INVALID_ARGUMENT,
-                     "[table][create][token-filter] "
-                     "nonexistent token filter: <%.*s>",
-                     token_filter_name_length, token_filter_name);
-    return GRN_FALSE;
-  }
-}
-
-static grn_bool
-grn_proc_table_set_token_filters_fill(grn_ctx *ctx,
-                                      grn_obj *token_filters,
-                                      grn_raw_string *token_filters_raw)
-{
-  const char *start, *current, *end;
-  const char *name_start, *name_end;
-  const char *last_name_end;
-
-  start = token_filters_raw->value;
-  end = start + token_filters_raw->length;
-  current = start;
-  name_start = NULL;
-  name_end = NULL;
-  last_name_end = start;
-  while (current < end) {
-    switch (current[0]) {
-    case ' ' :
-      if (name_start && !name_end) {
-        name_end = current;
-      }
-      break;
-    case ',' :
-      if (!name_start) {
-        goto break_loop;
-      }
-      if (!name_end) {
-        name_end = current;
-      }
-      if (!grn_proc_table_set_token_filters_put(ctx,
-                                                token_filters,
-                                                name_start,
-                                                name_end - name_start)) {
-        return GRN_FALSE;
-      }
-      last_name_end = name_end + 1;
-      name_start = NULL;
-      name_end = NULL;
-      break;
-    default :
-      if (!name_start) {
-        name_start = current;
-      }
-      break;
-    }
-    current++;
-  }
-
-break_loop:
-  if (!name_start) {
-    GRN_PLUGIN_ERROR(ctx,
-                     GRN_INVALID_ARGUMENT,
-                     "[table][create][token-filter] empty token filter name: "
-                     "<%.*s|%.*s|%.*s>",
-                     (int)(last_name_end - start), start,
-                     (int)(current - last_name_end), last_name_end,
-                     (int)(end - current), current);
-    return GRN_FALSE;
-  }
-
-  if (!name_end) {
-    name_end = current;
-  }
-  if (!grn_proc_table_set_token_filters_put(ctx,
-                                            token_filters,
-                                            name_start,
-                                            name_end - name_start)) {
-    return GRN_FALSE;
-  }
-
-  return GRN_TRUE;
-}
-
 grn_bool
 grn_proc_table_set_token_filters(grn_ctx *ctx,
                                  grn_obj *table,
@@ -176,14 +79,13 @@ grn_proc_table_set_token_filters(grn_ctx *ctx,
     return GRN_TRUE;
   }
 
-  GRN_PTR_INIT(&token_filters, GRN_OBJ_VECTOR, 0);
-  succeeded = grn_proc_table_set_token_filters_fill(ctx,
-                                                    &token_filters,
-                                                    token_filters_raw);
-  if (succeeded) {
-    grn_obj_set_info(ctx, table, GRN_INFO_TOKEN_FILTERS, &token_filters);
-  }
-  grn_obj_unlink(ctx, &token_filters);
+  GRN_TEXT_INIT(&token_filters, GRN_OBJ_DO_SHALLOW_COPY);
+  GRN_TEXT_SET(ctx,
+               &token_filters,
+               token_filters_raw->value,
+               token_filters_raw->length);
+  grn_obj_set_info(ctx, table, GRN_INFO_TOKEN_FILTERS, &token_filters);
+  GRN_OBJ_FIN(ctx, &token_filters);
 
   return succeeded;
 }
