@@ -1,6 +1,7 @@
 /* -*- c-basic-offset: 2 -*- */
 /*
   Copyright(C) 2009-2018 Brazil
+  Copyright(C) 2018 Kouhei Sutou <kou@clear-code.com>
 
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
@@ -28,6 +29,7 @@ grn_token_cursor_open_initialize_token_filters(grn_ctx *ctx,
 {
   grn_obj *token_filters = token_cursor->token_filter.objects;
   unsigned int i, n_token_filters;
+  grn_tokenizer_query *query = &(token_cursor->tokenizer.query);
 
   token_cursor->token_filter.data = NULL;
 
@@ -50,10 +52,16 @@ grn_token_cursor_open_initialize_token_filters(grn_ctx *ctx,
     grn_obj *token_filter_object = GRN_PTR_VALUE_AT(token_filters, i);
     grn_proc *token_filter = (grn_proc *)token_filter_object;
 
-    token_cursor->token_filter.data[i] =
-      token_filter->callbacks.token_filter.init(ctx,
-                                                token_cursor->table,
-                                                token_cursor->mode);
+    if (token_filter->callbacks.token_filter.init_query) {
+      grn_tokenizer_query_set_token_filter_index(ctx, query, i);
+      token_cursor->token_filter.data[i] =
+        token_filter->callbacks.token_filter.init_query(ctx, query);
+    } else {
+      token_cursor->token_filter.data[i] =
+        token_filter->callbacks.token_filter.init(ctx,
+                                                  token_cursor->table,
+                                                  token_cursor->mode);
+    }
   }
 }
 
@@ -62,6 +70,8 @@ grn_token_cursor_open(grn_ctx *ctx, grn_obj *table,
                       const char *str, size_t str_len,
                       grn_tokenize_mode mode, unsigned int flags)
 {
+  GRN_API_ENTER;
+
   grn_token_cursor *token_cursor;
   grn_encoding encoding;
   grn_obj *tokenizer;
@@ -69,7 +79,9 @@ grn_token_cursor_open(grn_ctx *ctx, grn_obj *table,
   grn_table_flags table_flags;
   if (grn_table_get_info(ctx, table, &table_flags, &encoding, &tokenizer,
                          NULL, &token_filters)) {
-    return NULL;
+    ERR(GRN_INVALID_ARGUMENT,
+        "[token-cursor][open] failed to get table information");
+    GRN_API_RETURN(NULL);
   }
   if (!(token_cursor = GRN_MALLOC(sizeof(grn_token_cursor)))) { return NULL; }
   token_cursor->table = table;
@@ -90,7 +102,6 @@ grn_token_cursor_open(grn_ctx *ctx, grn_obj *table,
   token_cursor->curr_size = 0;
   token_cursor->pos = -1;
   token_cursor->status = GRN_TOKEN_CURSOR_DOING;
-  token_cursor->force_prefix = GRN_FALSE;
   if (tokenizer) {
     grn_proc *tokenizer_proc = (grn_proc *)tokenizer;
     if (tokenizer_proc->callbacks.tokenizer.init) {
@@ -153,7 +164,7 @@ exit :
     grn_token_cursor_close(ctx, token_cursor);
     token_cursor = NULL;
   }
-  return token_cursor;
+  GRN_API_RETURN(token_cursor);
 }
 
 static int
@@ -164,6 +175,7 @@ grn_token_cursor_next_apply_token_filters(grn_ctx *ctx,
   unsigned int i, n_token_filters;
   grn_token *current_token = &(token_cursor->tokenizer.current_token);
   grn_token *next_token = &(token_cursor->tokenizer.next_token);
+  grn_tokenizer_query *query = &(token_cursor->tokenizer.query);
 
   if (token_filters) {
     n_token_filters = GRN_BULK_VSIZE(token_filters) / sizeof(grn_obj *);
@@ -178,6 +190,8 @@ grn_token_cursor_next_apply_token_filters(grn_ctx *ctx,
       grn_proc *token_filter = (grn_proc *)token_filter_object;
       void *data = token_cursor->token_filter.data[i];
 
+      grn_tokenizer_query_set_token_filter_index(ctx, query, i);
+
 #define SKIP_FLAGS                              \
       (GRN_TOKEN_SKIP |                         \
        GRN_TOKEN_SKIP_WITH_POSITION)
@@ -206,6 +220,8 @@ grn_token_cursor_next_apply_token_filters(grn_ctx *ctx,
 grn_id
 grn_token_cursor_next(grn_ctx *ctx, grn_token_cursor *token_cursor)
 {
+  GRN_API_ENTER;
+
   int status;
   grn_id tid = GRN_ID_NIL;
   grn_obj *table = token_cursor->table;
@@ -242,7 +258,6 @@ grn_token_cursor_next(grn_ctx *ctx, grn_token_cursor *token_cursor)
          (token_cursor->mode == GRN_TOKENIZE_GET &&
           (status & GRN_TOKEN_REACH_END)))
         ? GRN_TOKEN_CURSOR_DONE : GRN_TOKEN_CURSOR_DOING;
-      token_cursor->force_prefix = GRN_FALSE;
 #define SKIP_FLAGS \
       (GRN_TOKEN_SKIP | GRN_TOKEN_SKIP_WITH_POSITION)
       if (status & SKIP_FLAGS) {
@@ -258,7 +273,7 @@ grn_token_cursor_next(grn_ctx *ctx, grn_token_cursor *token_cursor)
       }
 #undef SKIP_FLAGS
       if (status & GRN_TOKEN_FORCE_PREFIX) {
-        token_cursor->force_prefix = GRN_TRUE;
+        grn_token_set_force_prefix_search(ctx, current_token, GRN_TRUE);
       }
       if (token_cursor->curr_size == 0) {
         if (token_cursor->status != GRN_TOKEN_CURSOR_DONE) {
@@ -291,7 +306,7 @@ grn_token_cursor_next(grn_ctx *ctx, grn_token_cursor *token_cursor)
           }
         } else {
           if (status & GRN_TOKEN_REACH_END) {
-            token_cursor->force_prefix = GRN_TRUE;
+            grn_token_set_force_prefix_search(ctx, current_token, GRN_TRUE);
           }
         }
       }
@@ -360,9 +375,22 @@ grn_token_cursor_next(grn_ctx *ctx, grn_token_cursor *token_cursor)
       token_cursor->status = GRN_TOKEN_CURSOR_NOT_FOUND;
     }
     token_cursor->pos++;
+    grn_token_set_position(ctx, current_token, token_cursor->pos);
     break;
   }
-  return tid;
+  GRN_API_RETURN(tid);
+}
+
+grn_token_cursor_status
+grn_token_cursor_get_status(grn_ctx *ctx, grn_token_cursor *token_cursor)
+{
+  GRN_API_ENTER;
+  if (!token_cursor) {
+    ERR(GRN_INVALID_ARGUMENT,
+        "[token-cursor][status] token cursor must not NULL");
+    GRN_API_RETURN(GRN_TOKEN_CURSOR_DONE);
+  }
+  GRN_API_RETURN(token_cursor->status);
 }
 
 static void
@@ -371,6 +399,7 @@ grn_token_cursor_close_token_filters(grn_ctx *ctx,
 {
   grn_obj *token_filters = token_cursor->token_filter.objects;
   unsigned int i, n_token_filters;
+  grn_tokenizer_query *query = &(token_cursor->tokenizer.query);
 
   if (!token_cursor->token_filter.data) {
     return;
@@ -391,6 +420,7 @@ grn_token_cursor_close_token_filters(grn_ctx *ctx,
     grn_proc *token_filter = (grn_proc *)token_filter_object;
     void *data = token_cursor->token_filter.data[i];
 
+    grn_tokenizer_query_set_token_filter_index(ctx, query, i);
     token_filter->callbacks.token_filter.fin(ctx, data);
   }
   GRN_FREE(token_cursor->token_filter.data);
@@ -399,32 +429,34 @@ grn_token_cursor_close_token_filters(grn_ctx *ctx,
 grn_rc
 grn_token_cursor_close(grn_ctx *ctx, grn_token_cursor *token_cursor)
 {
-  if (token_cursor) {
-    if (token_cursor->tokenizer.object) {
-      grn_proc *tokenizer_proc = (grn_proc *)(token_cursor->tokenizer.object);
-      if (tokenizer_proc->callbacks.tokenizer.fin) {
-        void *user_data = token_cursor->tokenizer.user_data;
-        tokenizer_proc->callbacks.tokenizer.fin(ctx, user_data);
-      } else if (tokenizer_proc->funcs[PROC_FIN]) {
-        tokenizer_proc->funcs[PROC_FIN](ctx,
-                                        1,
-                                        &token_cursor->table,
-                                        &token_cursor->tokenizer.pctx.user_data);
-      }
-    }
-    grn_token_fin(ctx, &(token_cursor->tokenizer.current_token));
-    grn_token_fin(ctx, &(token_cursor->tokenizer.next_token));
-    grn_tokenizer_query_fin(ctx, &(token_cursor->tokenizer.query));
-    grn_token_cursor_close_token_filters(ctx, token_cursor);
-    GRN_FREE(token_cursor);
-    return GRN_SUCCESS;
-  } else {
-    return GRN_INVALID_ARGUMENT;
+  GRN_API_ENTER;
+  if (!token_cursor) {
+    GRN_API_RETURN(GRN_INVALID_ARGUMENT);
   }
+
+  if (token_cursor->tokenizer.object) {
+    grn_proc *tokenizer_proc = (grn_proc *)(token_cursor->tokenizer.object);
+    if (tokenizer_proc->callbacks.tokenizer.fin) {
+      void *user_data = token_cursor->tokenizer.user_data;
+      tokenizer_proc->callbacks.tokenizer.fin(ctx, user_data);
+    } else if (tokenizer_proc->funcs[PROC_FIN]) {
+      tokenizer_proc->funcs[PROC_FIN](ctx,
+                                      1,
+                                      &token_cursor->table,
+                                      &token_cursor->tokenizer.pctx.user_data);
+    }
+  }
+  grn_token_fin(ctx, &(token_cursor->tokenizer.current_token));
+  grn_token_fin(ctx, &(token_cursor->tokenizer.next_token));
+  grn_tokenizer_query_fin(ctx, &(token_cursor->tokenizer.query));
+  grn_token_cursor_close_token_filters(ctx, token_cursor);
+  GRN_FREE(token_cursor);
+  GRN_API_RETURN(GRN_SUCCESS);
 }
 
 grn_token *
 grn_token_cursor_get_token(grn_ctx *ctx, grn_token_cursor *token_cursor)
 {
-  return &(token_cursor->tokenizer.current_token);
+  GRN_API_ENTER;
+  GRN_API_RETURN(&(token_cursor->tokenizer.current_token));
 }
